package net.sergeych.parsec3

import io.ktor.client.*
import io.ktor.client.plugins.websocket.*
import io.ktor.websocket.*
import kotlinx.coroutines.CompletableDeferred
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.flow.StateFlow
import kotlinx.coroutines.launch
import net.sergeych.mp_logger.LogTag
import net.sergeych.mp_logger.info
import net.sergeych.mp_tools.globalLaunch

/**
 * Construct websocket-based client with client-side API (called _from server_). This form is universal
 * and basically is needed when client is accepting synchronous data calls, e.g. pushes, from the server.
 * There is a simpler constructor when it is not needed.
 *
 * @param url server url to connect to
 * @param api client api to implement in the builder
 * @param exceptionsRegistry the registry of supported exceptions that can be safely transmitted over the network
 * @param builder client side api builder, called over api instance. Here client can _implement_ the commands that server
 *          could call.
 */
class Parsec3WSClient<S : WithAdapter, H : CommandHost<S>>(
    val url: String,
    api: H,
    override val exceptionsRegistry: ExceptionsRegistry = ExceptionsRegistry(),
    builder: AdapterBuilder<S, H>.() -> Unit,
) : LogTag("P3WSC"), Parsec3Transport<S> {

    val builder = AdapterBuilder(api, exceptionsRegistry, builder)

    private val _connectionFlow = MutableStateFlow(false)
    private val closeFlow = MutableStateFlow(false)
    private val reconnectFlow = MutableStateFlow(false)
    override val connectedFlow: StateFlow<Boolean> = _connectionFlow


    init {
        start()
    }

    override fun close() {
        if (closeFlow.value == false) closeFlow.value = true
    }

    override fun reconnect() {
        reconnectFlow.value = true
    }

    var deferredAdapter = CompletableDeferred<Adapter<S>>()
        private set

    override suspend fun adapter(): Adapter<S> = deferredAdapter.await()

    fun start() {
        globalLaunch {
            while (closeFlow.value != true) {
                reconnectFlow.value = false
                client.webSocket(url) {
                    info { "Connected to $url" }
                    val a = builder.create { send(Frame.Binary(true, it)) }
                    _connectionFlow.value = true
                    launch { closeFlow.collect { if (it == true) close() } }
                    launch { reconnectFlow.collect { if (it == true) close() } }
                    deferredAdapter.complete(a)
                    for (f in incoming)
                        if (f is Frame.Binary) a.receiveFrame(f.data)
                    // when we leave connection will be closed. So far we do not close the adapter,
                    // should we?
                    _connectionFlow.value = false
                    deferredAdapter = CompletableDeferred()
                }
            }
        }
    }

    companion object {
        private val client = HttpClient {
            install(WebSockets) {
                // Configure WebSockets
            }
        }

        /**
         * Simplified client constructor for the case when client does not receives commands (e.g. pushes)
         * from the server.
         * @param url server url
         * @param exceptionsRegistry converter of exceptions that can be received from the remote.
         */
        operator fun invoke(url: String, exceptionsRegistry: ExceptionsRegistry = ExceptionsRegistry())
                : Parsec3WSClient<WithAdapter, CommandHost<WithAdapter>> {
            return Parsec3WSClient(url, CommandHost<WithAdapter>(), exceptionsRegistry) {}
        }

        @Suppress("unused")
        fun <S: WithAdapter>withSession(url: String): Parsec3WSClient<S, CommandHost<S>> {
            return Parsec3WSClient(url, CommandHost(), ExceptionsRegistry()) {}
        }

    }
}
